/**************************************************************************
 *
 * FILE  cdis.c
 * Copyright (c) 2018, 2019 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   cbm binary disassembler
 *
 ******/
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "cbm_mem.h"
#include "global.h"
#include "opcodes.h"
#include "output.h"
#include "params.h"
#include "stream_io.h"

#define PROGRAM "cdis"

const char program_g[] = PROGRAM;
const int exit_code_fail_g = 1;

typedef struct {
    int longaddr;
    cbm_ptr_t loadaddr;
    int rawfile;
    int undoc;
} Options;

static Options opt;

static void dump_file(FILE *fp);

int main(int argc, char *argv[])
{
    int c;

    /* defaults */
    verbose_g = 0;
    debug_g = 0;
    opt.longaddr = 0;
    opt.loadaddr = CBM_PTR_UNDEFINED;
    opt.rawfile = 0;
    opt.undoc = 0;

    /*
     * scan for valid options
     */
    while ( c = getopt(argc, argv, "bl:LudvVh"), c != EOF ) {
        switch (c) {

	/* a missing parameter */
	case ':':
	/* an illegal option */
	case '?':
	    exit(1);

	/* set verbose */
	case 'v':
	    verbose_g = 1;
	    break;

	/* set debug */
	case 'd':
	    debug_g = 1;
	    break;

	/* print version */
	case 'V':
	    fprintf(stdout, PROGRAM " (" PACKAGE ") " VERSION "\n");
	    exit(0);

	/* print help */
	case 'h':
	    fprintf(stderr,
PROGRAM " (" PACKAGE ") " VERSION ": cbm binary disassembler\n"
"Copyright (c) 2018, 2019 Daniel Kahlin <daniel@kahlin.net>\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
"\n"
"usage: " PROGRAM " [-b][-l<addr>][-L][-u][-v][-d][-h][-V] <file>...\n"
"\n"
"Valid options:\n"
"    -b              file has no load address\n"
"    -l              override load address\n"
"    -L              24-bit addressing (default: 16-bit)\n"
"    -u              disassemble undocumented opcodes\n"
"    -v              be verbose\n"
"    -d              debugging output\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
	    );
	    exit(0);

	/* set raw mode */
	case 'b':
	    opt.rawfile = 1;
	    break;

	/* set Load address */
	case 'l':
	    opt.loadaddr = parse_value(optarg);
	    break;

	/* set Long addressing */
	case 'L':
	    opt.longaddr = 1;
	    break;

	/* set undocmented mode */
	case 'u':
	    opt.undoc = 1;
	    break;

	/* default behavior */
	default:
	    break;
	}
    }

    /*
     * optind now points at the first non option argument
     */
    int num_files = argc - optind;
    char **srcnames = &argv[optind];

    /* if no arguments, dump from stdin */
    if (num_files == 0) {
	static char *stdinname = "-";
	srcnames = &stdinname;
	num_files++;
    }

    /* if one or more arguments, dump from the listed files */
    int ret = 0;
    int i;
    for (i = 0; i < num_files; i++) {
	int c;
	Stream *st;
	char *srcname;
	srcname = srcnames[i];

	st = stream_open_warn(srcname, "rb", 0);
	if (!st) {
	    ret = 1;
	    continue;
	}

	if (num_files > 1) {
	    printf("--- %s ---\n", st->print_name);
	}

	dump_file(st->fp);

	c = stream_close_warn(st);
	if (c == EOF) {
	    ret = 1;
	    continue;
	}

    }

    exit(ret);
}


static cbm_ptr_t inc_addr(cbm_ptr_t ad)
{
    ad++;
    if (!opt.longaddr) {
	ad &= 0xffff;
    } else {
	ad &= 0xffffff;
    }

    return ad;
}

static int is_invalid(uint8_t op)
{
    if ( !opt.undoc && (tab_umode[op] & OP_UND) ) {
	return 1;
    }
    return 0;
}

static int get_instr_size(uint8_t op)
{
    if (is_invalid(op)) {
	return 1;
    }
    return tab_size[op];
}

static const char *get_instr_mne(uint8_t op)
{
    if (is_invalid(op)) {
	return "???";
    }

    return tab_mne[op];
}

static int get_instr_amode(uint8_t op)
{
    if (is_invalid(op)) {
	return OP_IMP;
    }

    return tab_amode[op];
}


static void disasm_line(Dumper *dmp, cbm_ptr_t addr, uint8_t *buf, int len)
{
    int i;
    int l;
    uint8_t op;

    if (dmp->long_addr) {
	printf("%06X ", addr);
    } else {
	printf("%04X ", addr);
    }


    op = buf[0];
    l = get_instr_size(op);

    /* check if the full instruction fits before the end of file */
    if (len < l) {
	for (i = 0; i < len; i++) {
	    printf(" %02X", buf[i]);
	}
	printf("\n");
	return;
    }

    for (i = 0; i < 3; i++) {
	if (i < l) {
	    printf(" %02X", buf[i]);
	} else {
	    printf("   ");
	}
    }
    char mne[4];
    const char *orig = get_instr_mne(op);
    for (i = 0; i < 3; i++) {
	mne[i] = toupper(orig[i]);
    }
    mne[3] = 0;

    printf("  %s", mne);

    switch (get_instr_amode(op)) {
    case OP_IMP:
	break;
    case OP_IMM:
	printf(" #$%02X", buf[1]);
	break;
    case OP_ZP:
	printf(" $%02X", buf[1]);
	break;
    case OP_ZPX:
	printf(" $%02X,X", buf[1]);
	break;
    case OP_ZPY:
	printf(" $%02X,Y", buf[1]);
	break;
    case OP_ABS:
	printf(" $%02X%02X", buf[2], buf[1]);
	break;
    case OP_ABX:
	printf(" $%02X%02X,X", buf[2], buf[1]);
	break;
    case OP_ABY:
	printf(" $%02X%02X,Y", buf[2], buf[1]);
	break;
    case OP_IZX:
	printf(" ($%02X,X)", buf[1]);
	break;
    case OP_IZY:
	printf(" ($%02X),Y", buf[1]);
	break;
    case OP_REL:
	printf(" $%04X", (uint16_t)(addr + (int8_t)buf[1] + 2));
	break;
    case OP_IND:
	printf(" ($%02X%02X)", buf[2], buf[1]);
	break;
    default:
	break;
    }

    printf("\n");
}


static void flush_disasm(Dumper *dmp)
{
    if (dmp->len) {
	disasm_line(dmp, dmp->addr, dmp->buf, dmp->len);
	dmp->len = 0;
    }
}


static void disasm_data(Dumper *dmp, cbm_ptr_t addr, uint8_t data)
{
    if (dmp->len == 0) {
	dmp->addr = addr;
    }

    dmp->buf[dmp->len++] = data;

    if (dmp->len && dmp->len == get_instr_size(dmp->buf[0])) {
	flush_disasm(dmp);
    }
}



static void dump_file(FILE *fp)
{
    cbm_ptr_t ad;
    cbm_range_t rng = CBM_RANGE_UNDEFINED;
    int nread = 0;
    Dumper dmp;

    init_dumper(&dmp);
    dmp.long_addr = opt.longaddr;

    if (!opt.rawfile) {
	ad = fgetc(fp) + (fgetc(fp) << 8);
	nread += 2;
    } else {
	ad = 0x0000;
    }
    if (is_defined_addr(opt.loadaddr)) {
	ad = opt.loadaddr;
    }

    /* initial empty range */
    rng.sa = ad;
    rng.ea = ad;

    int c;
    while (c = fgetc(fp), c != EOF) {
	nread++;

	disasm_data(&dmp, ad, c);
	update_range(&rng, ad);

	/* next address */
	ad = inc_addr(ad);

    }
    flush_disasm(&dmp);


    int nblks = (nread + 253) / 254;
    if (opt.longaddr) {
	rng.sa &= 0xffffff;
	rng.ea &= 0xffffff;
	printf("disassembled $%06X-$%06X (%d %s, %d %s)\n", rng.sa, rng.ea, nread, nread==1?"byte":"bytes", nblks, nblks==1?"block":"blocks");
    } else {
	rng.sa &= 0xffff;
	rng.ea &= 0xffff;
	printf("disassembled $%04X-$%04X (%d %s, %d %s)\n", rng.sa, rng.ea, nread, nread==1?"byte":"bytes", nblks, nblks==1?"block":"blocks");
    }
}


/* eof */
