/**************************************************************************
 *
 * FILE  stream_io.c
 * Copyright (c) 2019 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   common I/O functions
 *
 ******/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "compat.h"
#include "global.h"
#include "stream_io.h"


Stream *stream_open(const char *srcname, const char *mode, const char **print_name)
{
    Stream *st;
    st = malloc(sizeof(Stream));
    if (!st) {
	panic("couldn't allocate mem");
    }

    st->bin_mode = 1;
    if (mode[0] == 'w') {
	st->write_mode = 1;
    } else {
	st->write_mode = 0;
    }

    if (!strcmp(srcname, "-")) {
	if (st->write_mode) {
	    st->print_name = "<stdout>";
	    st->fp = stdout;
	} else {
	    st->print_name = "<stdin>";
	    st->fp = stdin;
	}
	if (st->bin_mode) {
	    set_binmode(st->fp);
	}
	*print_name = st->print_name;
    } else {
	st->print_name = srcname;
	*print_name = st->print_name;

	st->fp = fopen(srcname, mode);
	if (!st->fp) {
	    free(st);
	    return 0;
	}

	if (!st->write_mode) {
	    /* force any pending error by actually reading from the file */
	    fgetc(st->fp);
	    if (ferror(st->fp)) {
		free(st);
		return 0;
	    }
	    fseek(st->fp, 0, SEEK_SET);
	}
    }

    return st;
}


int stream_close(Stream *st)
{
    if (ferror(st->fp)) {
	if (st->fp == stdin || st->fp == stdout) {
	    if (st->fp == stdout) {
		fflush(st->fp);
	    }
	    if (st->bin_mode) {
		set_textmode(st->fp);
	    }
	}
	free(st);
	return EOF;
    }

    if (st->fp == stdin || st->fp == stdout) {
	if (st->fp == stdout) {
	    fflush(st->fp);
	}
	if (st->bin_mode) {
	    set_textmode(st->fp);
	}
    } else {
	int c = fclose(st->fp);
	if (c == EOF) {
	    free(st);
	    return EOF;
	}
    }

    free(st);
    return 0;
}


Stream *stream_open_warn(const char *name, const char *mode, const char **print_name)
{
    const char *tmp;
    if (!print_name) {
	print_name = &tmp;
    }

    Stream *st = stream_open(name, mode, print_name);
    if (!st) {
	fprintf(stderr, "%s: %s: %s\n", program_g, *print_name, strerror(errno));
	return 0;
    }

    return st;
}

int stream_close_warn(Stream *st)
{
    const char *print_name = st->print_name;

    int c = stream_close(st);
    if (c == EOF) {
	fprintf(stderr, "%s: %s: %s\n", program_g, print_name, strerror(errno));
	return EOF;
    }

    return c;
}

Stream *stream_open_fail(const char *name, const char *mode, const char **print_name)
{
    Stream *st = stream_open_warn(name, mode, print_name);
    if (!st) {
	exit(exit_code_fail_g);
    }

    return st;
}

int stream_close_fail(Stream *st)
{
    int c = stream_close_warn(st);
    if (c == EOF) {
	exit(exit_code_fail_g);
    }

    return c;
}


/* eof */
